import os
import subprocess
from typing import Set

from ..core.log import log
from .path import get_relative_path


def get_root_directory(path: str = '') -> str:
    """
    :raises: CalledProcessError
    """
    command = ['git']
    if path:
        command.extend(['-C', path])

    command.extend(['rev-parse', '--show-toplevel'])
    return subprocess.check_output(command).decode('utf-8').strip()  # noqa: S603


def get_tracked_files(root: str) -> Set[str]:
    """Parsing .gitignore rules is hard.

    However, a way we can get around this problem by just listing all
    currently tracked git files, and start our search from there.
    After all, if it isn't in the git repo, we're not concerned about
    it, because secrets aren't being entered in a shared place.

    :raises: CalledProcessError
    """
    output = set()
    try:
        files = subprocess.check_output(
            ['git', '-C', root, 'ls-files'],  # noqa: S603,S607
            stderr=subprocess.DEVNULL,
        )

        for filename in files.decode('utf-8').splitlines():
            path = get_relative_path(root, os.path.join(root, filename))
            if path:
                output.add(path)

    except subprocess.CalledProcessError:
        pass
    except FileNotFoundError:   # pragma: no cover
        log.warning('Unable to find `git` in PATH, and therefore, unable to get tracked files.')

    return output


def get_changed_but_unstaged_files() -> Set[str]:
    try:
        files = subprocess.check_output('git diff --name-only'.split()).decode().splitlines()  # noqa: S603
    except subprocess.CalledProcessError:   # pragma: no cover
        # Since we don't pipe stderr, we get free logging through git.
        raise ValueError

    return set(files)



import re

from .base import RegexBasedDetector


class ArtifactoryDetector(RegexBasedDetector):
    """Scans for Artifactory credentials."""
    secret_type = 'Artifactory Credentials'

    denylist = [
        # Artifactory tokens begin with AKC
        # API token:
        re.compile(r'(?:\s|=|:|"|^)AKC[a-zA-Z0-9]{10,}(?:\s|"|$)'),
        # Artifactory encrypted passwords begin with AP[A-Z]
        # Password:
        re.compile(r'(?<!AAAA[_\-\w]{7})(?:\s|=|:|"|^)AP[\dABCDEF][a-zA-Z0-9]{8,}(?:\s|"|$)'),
    ]



import argparse
import os

from ...settings import default_settings
from ...settings import get_settings


def valid_path(path: str) -> str:
    if not os.path.isfile(path):
        raise argparse.ArgumentTypeError(
            f'Invalid path: {path}',
        )

    return path


def initialize_plugin_settings(args: argparse.Namespace) -> None:  # noqa: ARG001
    """
    This is a stand-in function, which should be replaced if baseline options are used.
    This ensures that our global settings object is initialized to a minimal state
    (all built-in plugins, default options)
    """
    # This is a sanity check to ensure we don't override any current settings.
    if get_settings().plugins:
        return

    # We initialize the `settings` variable here, but we can't save it to the global object
    # yet, since the contextmanager will revert those changes. As such, we quit the context
    # first, then set it to the global namespace.
    with default_settings() as settings:
        pass

    get_settings().set(settings)



import os
import re
from functools import lru_cache
from typing import Dict
from typing import Iterable
from typing import List
from typing import Pattern
from typing import Tuple

from ..util.code_snippet import CodeSnippet


def is_line_allowlisted(filename: str, line: str, context: CodeSnippet) -> bool:
    for payload, regexes in zip(
        [line, context.previous_line],
        _get_allowlist_regexes_for_file(filename),
    ):
        for regex in regexes:
            if regex.search(payload):
                return True

    return False


@lru_cache(maxsize=1)
def _get_file_to_index_dict() -> Dict[str, int]:
    # Add to this mapping (and ALLOWLIST_REGEXES if applicable) lazily,
    # as more language specific file parsers are implemented.
    # Discussion: https://github.com/Yelp/detect-secrets/pull/105
    return {
        'yaml': 0,
    }


@lru_cache(maxsize=1)
def _get_comment_tuples() -> List[Tuple[str, str]]:
    return [
        ('#', ''),                    # e.g. python or yaml
        ('//', ''),                   # e.g. golang
        (r'/\*', r' *\*/'),           # e.g. c
        ('\'', ''),                   # e.g. visual basic .net
        ('--', ''),                   # e.g. sql
        (r'<!--[# \t]*?', ' *?-->'),  # e.g. xml
        # many other inline comment syntaxes are not included,
        # because we want to be performant for
        # any(regex.search(line) for regex in ALLOWLIST_REGEXES)
        # calls. of course, this won't be a concern if detect-secrets
        # switches over to implementing file plugins for each supported
        # filetype.
    ]


def _get_allowlist_regexes_for_file(filename: str) -> Iterable[List[Pattern]]:
    comment_tuples = _get_comment_tuples()

    _, ext = os.path.splitext(filename)
    if ext[1:] in _get_file_to_index_dict():
        comment_tuples = [comment_tuples[_get_file_to_index_dict()[ext[1:]]]]

    yield [
        get_allowlist_regexes(comment_tuple=t, nextline=False)
        for t in comment_tuples
    ]
    yield [
        get_allowlist_regexes(comment_tuple=t, nextline=True)
        for t in comment_tuples
    ]


# Note: Cache size should be 2x the number of comment types
@lru_cache(maxsize=12)
def get_allowlist_regexes(comment_tuple: Tuple[str, str], nextline: bool) -> Pattern:
    start = comment_tuple[0]
    end = comment_tuple[1]
    return re.compile(
        r'{}[ \t]*{} *pragma: ?{}{}[ -]secret.*?{}[ \t]*$'.format(
            # Note: No text can precede a nextline pragma, this prevents obscuring what is allowed
            # For instance, we want to prevent the following case from working:
            #     foo = 'bar' # pragma: allowlist nextline secret
            #     pass = 'hunter2'
            r'^' if nextline else '',
            start,
            # Note: Always use allowlist, whitelist will be deprecated in the future
            r'allowlist' if nextline else r'(allow|white)list',
            r'[ -]nextline' if nextline else '',
            end,
        ),
    )



from __future__ import annotations

import argparse
import os
from typing import cast
from typing import Iterable

from .. import plugins
from ...exceptions import InvalidFile
from ...settings import get_settings
from ..plugins.util import get_mapping_from_secret_type_to_class


def add_plugin_options(parent: argparse.ArgumentParser) -> None:
    parser = parent.add_argument_group(
        title='plugin options',
        description=(
            'Configure settings for each secret scanning '
            'ruleset. By default, all plugins are enabled '
            'unless explicitly disabled.'
        ),
    )

    parser.add_argument(
        '--list-all-plugins',
        action='store_true',
        help='Lists all plugins that will be used for the scan.',
    )

    _add_custom_plugins(parser)
    _add_custom_limits(parser)
    _add_disable_flag(parser)


def _add_custom_plugins(parser: argparse._ArgumentGroup) -> None:
    def valid_looking_paths(path: str) -> str:
        # We can verify whether these files are valid at post-processing.
        # TODO: support directories
        # TODO: support selecting specific classes in files.
        # TODO: do we also want to overload this and allow specific selection of plugins for
        # baselines that don't currently use those plugins?
        if not os.path.isfile(path):
            raise argparse.ArgumentTypeError(f'{path} is not a valid file.')

        return path

    parser.add_argument(
        '-p',
        '--plugin',
        type=valid_looking_paths,
        nargs=1,
        action='append',        # so we can support multiple flags with same value
        help='Specify path to custom secret detector plugin.',
    )


def _add_custom_limits(parser: argparse._ArgumentGroup) -> None:
    def minmax_type(string: str) -> float:
        value = float(string)
        if value < 0 or value > 8:
            raise argparse.ArgumentTypeError(
                f'{string} must be between 0.0 and 8.0',
            )

        return value

    high_entropy_help_text = (
        'Sets the entropy limit for high entropy strings. '
        'Value must be between 0.0 and 8.0,'
    )

    # NOTE: This doesn't have explicit default values since we want to be able to determine
    # the difference between a value set by default, and an explicitly set value (which happens
    # to be the same as the default value). This distinction plays an important role when doing
    # precedence calculation (default value < baseline config < CLI explicit value)
    parser.add_argument(
        '--base64-limit',
        type=minmax_type,
        nargs='?',
        help=high_entropy_help_text + ' defaults to 4.5.',
    )
    parser.add_argument(
        '--hex-limit',
        type=minmax_type,
        nargs='?',
        help=high_entropy_help_text + ' defaults to 3.0.',
    )


def _add_disable_flag(parser: argparse._ArgumentGroup) -> None:
    def valid_plugin_name(string: str) -> str:
        valid_plugin_names: set[str] = {
            item.__name__
            for item in get_mapping_from_secret_type_to_class().values()
        }

        if string not in valid_plugin_names:
            raise argparse.ArgumentTypeError(f'Invalid plugin classname: {string}')

        return string

    parser.add_argument(
        '--disable-plugin',
        type=valid_plugin_name,
        nargs=1,
        action='append',        # so we can support multiple flags with the same value
        help='Plugin class names to disable. e.g. Base64HighEntropyString',
    )


def parse_args(args: argparse.Namespace) -> None:
    if args.disable_plugin:
        # Flatten entry for easier parsing.
        args.disable_plugin = {entry for item in args.disable_plugin for entry in item}
        get_settings().disable_plugins(*args.disable_plugin)

    # By the time the code reaches here, the baseline logic will have populated an initial
    # state for settings. Therefore, this will override whatever state that is currently registered.
    #
    # Default values will be applied at the plugin level.
    if args.base64_limit:
        get_settings().plugins['Base64HighEntropyString']['limit'] = args.base64_limit

    if args.hex_limit:
        get_settings().plugins['HexHighEntropyString']['limit'] = args.hex_limit

    if args.plugin:
        # Flatten entry for easier parsing.
        args.plugin = [entry for item in args.plugin for entry in item]

        for filename in args.plugin:
            # NOTE: Technically, we could just configure the settings, and have
            # `detect_secrets.core.plugins.util.get_mapping_from_secret_type_to_class`
            # to initialize them. However, if it's in the baseline / settings, we can
            # assume it works -- therefore, let's initialize it to discover any errors early
            # on, before storing it in settings.
            try:
                custom_plugins = cast(Iterable, plugins.initialize.from_file(filename))
            except InvalidFile:
                raise argparse.ArgumentTypeError(f'Cannot load plugins from {filename}.')

            get_settings().configure_plugins([
                {
                    'name': item.__name__,
                    'path': f'file://{os.path.abspath(filename)}',
                }
                for item in custom_plugins
            ])



"""
This plugin searches for Twilio API keys
"""
import re

from .base import RegexBasedDetector


class TwilioKeyDetector(RegexBasedDetector):
    """Scans for Twilio API keys."""
    secret_type = 'Twilio API Key'

    denylist = [
        # Account SID (ACxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx)
        re.compile(r'AC[a-z0-9]{32}'),

        # Auth token (SKxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx)
        re.compile(r'SK[a-z0-9]{32}'),
    ]



from __future__ import annotations

import os
from functools import lru_cache
from typing import cast
from typing import TYPE_CHECKING

import requests

from ..constants import VerifiedResult
from ..settings import get_settings
from ..util.code_snippet import CodeSnippet
from ..util.inject import call_function_with_arguments
from .util import get_caller_path

if TYPE_CHECKING:
    from detect_secrets.plugins.base import BasePlugin


def is_invalid_file(filename: str) -> bool:
    return not os.path.isfile(filename)


def is_baseline_file(filename: str) -> bool:
    return os.path.basename(filename) == _get_baseline_filename()


@lru_cache(maxsize=1)
def _get_baseline_filename() -> str:
    path = get_caller_path(offset=1)
    return cast(str, get_settings().filters[path]['filename'])


def is_ignored_due_to_verification_policies(
    secret: str,
    plugin: BasePlugin,
    context: CodeSnippet,
) -> bool:
    """
    Valid policies include:
        - Only VERIFIED_TRUE
        - Can be UNVERIFIED or VERIFIED_TRUE
        - Disabled check.

    There's no such thing as "only verified false", because if you're going to verify
    something, and it's verified false, why are you still including it as a valid secret?
    """
    try:
        verify_result = call_function_with_arguments(
            plugin.verify,
            secret=secret,
            context=context,
        )
    except requests.exceptions.RequestException:
        verify_result = VerifiedResult.UNVERIFIED

    if not verify_result:
        return False

    if verify_result.value < _get_verification_policy().value:
        return True

    return False


@lru_cache(maxsize=1)
def _get_verification_policy() -> VerifiedResult:
    path = get_caller_path(offset=1)
    return VerifiedResult(get_settings().filters[path]['min_level'])



import argparse
import json
import os
import sys
import textwrap
from typing import List
from typing import Optional

from detect_secrets.__version__ import VERSION
from detect_secrets.core import baseline
from detect_secrets.core.log import log
from detect_secrets.core.secrets_collection import SecretsCollection
from detect_secrets.core.usage import ParserBuilder
from detect_secrets.util import color
from detect_secrets.util import git


def main(argv: Optional[List[str]] = None) -> int:
    try:
        args = parse_args(argv)
    except ValueError:
        return 1

    if args.verbose:    # pragma: no cover
        log.set_debug_level(args.verbose)

    # Find all secrets in files to be committed
    secrets = SecretsCollection()
    for filename in args.filenames:
        secrets.scan_file(filename)

    new_secrets = secrets
    if args.baseline:
        new_secrets = secrets - args.baseline

    if new_secrets:
        if args.json:
            print(json.dumps(baseline.format_for_output(new_secrets), indent=2))
        else:
            pretty_print_diagnostics(new_secrets)
        return 1

    if not args.baseline:
        return 0

    # Only attempt baseline modifications if we don't find any new secrets.
    is_modified = should_update_baseline(
        args.baseline,
        scanned_results=secrets,
        filelist=args.filenames,
        baseline_version=args.baseline_version,
    )

    if is_modified:
        if args.baseline_version != VERSION:
            with open(args.baseline_filename) as f:
                old_baseline = json.loads(f.read())

            # Override the results, because this has been updated in `should_update_baseline`.
            old_baseline['results'] = args.baseline.json()

            args.baseline = baseline.upgrade(old_baseline)

        baseline.save_to_file(args.baseline, filename=args.baseline_filename)
        print(
            'The baseline file was updated.\n'
            'Probably to keep line numbers of secrets up-to-date.\n'
            'Please `git add {}`, thank you.\n\n'.format(args.baseline_filename),
        )
        return 3

    return 0


def parse_args(argv: Optional[List[str]] = None) -> argparse.Namespace:
    """
    :raises: ValueError
    """
    output = ParserBuilder().add_pre_commit_arguments().parse_args(argv)
    if output.baseline:
        raise_exception_if_baseline_file_is_unstaged(output.baseline_filename)

    return output


def raise_exception_if_baseline_file_is_unstaged(filename: str) -> None:
    """
    We want to make sure that if there are changes to the baseline
    file, they will be included in the commit. This way, we can keep
    our baselines up-to-date.

    :raises: ValueError
    """
    if filename in git.get_changed_but_unstaged_files():
        print(
            f'Your baseline file ({filename}) is unstaged.\n'
            f'`git add {filename}` to fix this.',
        )
        raise ValueError


def should_update_baseline(
    secrets: SecretsCollection,
    scanned_results: SecretsCollection,
    filelist: List[str],
    baseline_version: str,
) -> bool:
    """
    :returns: True if changes occurred.
    """
    original = SecretsCollection.load_from_baseline({'results': secrets.json()})

    secrets.trim(scanned_results=scanned_results, filelist=filelist)

    if baseline_version != VERSION:
        return True

    if not secrets.exactly_equals(original):
        return True

    return False


def pretty_print_diagnostics(secrets: SecretsCollection, width: int = 80) -> None:
    # Header
    print(
        textwrap.fill(
            color.colorize(
                'ERROR: Potential secrets about to be committed to git repo!',
                color.AnsiColor.RED,
            ),
            width=width,
        ),
    )
    print()

    # Display found secrets
    for _, secret in secrets:
        print(secret)

    # Display mitigation suggestions
    print('Possible mitigations:')
    wrapper = textwrap.TextWrapper(
        initial_indent='  - ',
        subsequent_indent='    ',
        width=width,
    )
    for suggestion in [
        'For information about putting your secrets in a safer place, '
        f'please ask {os.environ.get("DETECT_SECRETS_SECURITY_TEAM", "in #security")}',

        'Mark false positives with an inline '
        f'`{color.colorize("pragma: allowlist secret", color.AnsiColor.BOLD)}` comment',
    ]:
        print(wrapper.fill(suggestion))

    print()
    print(
        textwrap.fill(
            'If a secret has already been committed, visit '
            'https://help.github.com/articles/removing-sensitive-data-from-a-repository',
            width=width,
        ),
    )


if __name__ == '__main__':
    sys.exit(main())



import re
from base64 import b64encode

import requests

from ..constants import VerifiedResult
from .base import RegexBasedDetector


class StripeDetector(RegexBasedDetector):
    """Scans for Stripe keys."""
    secret_type = 'Stripe Access Key'

    denylist = (
        # Stripe standard keys begin with sk_live and restricted with rk_live
        re.compile(r'(?:r|s)k_live_[0-9a-zA-Z]{24}'),
    )

    def verify(self, secret: str) -> VerifiedResult:  # pragma: no cover
        response = requests.get(
            'https://api.stripe.com/v1/charges',
            headers={
                'Authorization': b'Basic ' + b64encode(
                    '{}:'.format(secret).encode('utf-8'),
                ),
            },
        )

        if response.status_code == 200:
            return VerifiedResult.VERIFIED_TRUE

        # Restricted keys may be limited to certain endpoints
        if secret.startswith('rk_live'):
            return VerifiedResult.UNVERIFIED

        return VerifiedResult.VERIFIED_FALSE



"""
The audit module allows analysts to easily scan the baseline, and verify whether
the secrets flagged are actually secrets.
"""
from . import io
from ..core import baseline
from ..custom_types import SecretContext
from ..exceptions import NoLineNumberError
from ..exceptions import SecretNotFoundOnSpecifiedLineError
from ..util.code_snippet import get_code_snippet
from .common import get_baseline_from_file
from .common import get_raw_secret_from_file
from .common import open_file
from .iterator import BidirectionalIterator
from .iterator import get_secret_iterator


def audit_baseline(filename: str) -> None:
    """
    :raises: InvalidBaselineError
    """
    secrets = get_baseline_from_file(filename)

    secrets.trim()
    if _classify_secrets(get_secret_iterator(secrets)):
        io.print_message('Saving progress...')
        baseline.save_to_file(secrets, filename)


def _classify_secrets(iterator: BidirectionalIterator) -> bool:
    """
    :returns: True if changes were made.
    """
    # NOTE: Technically, this is a conservative estimate. If an entry was changed to the same
    # value, we would return True as well.
    has_changes = False

    for secret in iterator:
        try:
            secret.secret_value = get_raw_secret_from_file(secret)
            io.clear_screen()
            io.print_context(
                SecretContext(
                    current_index=iterator.index + 1,
                    num_total_secrets=len(iterator.collection),
                    secret=secret,
                    snippet=get_code_snippet(
                        lines=open_file(secret.filename).raw_lines,
                        line_number=secret.line_number,
                    ),
                ),
            )

            decision = io.get_user_decision(can_step_back=iterator.can_step_back())
        except SecretNotFoundOnSpecifiedLineError as e:
            io.clear_screen()
            io.print_secret_not_found(
                SecretContext(
                    current_index=iterator.index + 1,
                    num_total_secrets=len(iterator.collection),
                    secret=secret,
                    error=e,
                ),
            )

            decision = io.get_user_decision(
                prompt_secret_decision=False,
                can_step_back=iterator.can_step_back(),
            )
        except NoLineNumberError as e:
            io.print_error(str(e))
            break

        if decision == io.InputOptions.QUIT:
            io.print_message('Quitting...')
            break

        if decision == io.InputOptions.BACK:
            iterator.step_back_on_next_iteration()

        # The question asked is: "Should this be committed to the repository?"
        elif decision == io.InputOptions.NO:
            secret.is_secret = True
            has_changes = True
        elif decision == io.InputOptions.YES:
            secret.is_secret = False
            has_changes = True
        elif decision == io.InputOptions.SKIP and secret.is_secret is not None:
            # This handles the case of back-stepping to clear a mistake.
            # This is not triggered for pre-labelled secrets, as pre-labelled secrets will be
            # excluded from this iterator.
            secret.is_secret = None
            has_changes = True

    return has_changes



import re

from .base import RegexBasedDetector


# This list is derived from RFC 3986 Section 2.2.
#
# We don't expect any of these delimiter characters to appear in
# the username/password component of the URL, seeing that this would probably
# result in an unexpected URL parsing (and probably won't even work).
RESERVED_CHARACTERS = ':/?#[]@'
SUB_DELIMITER_CHARACTERS = '!$&\'()*+,;='


class BasicAuthDetector(RegexBasedDetector):
    """Scans for Basic Auth formatted URIs."""
    secret_type = 'Basic Auth Credentials'

    denylist = [
        re.compile(
            r'://[^{}\s]+:([^{}\s]+)@'.format(
                re.escape(RESERVED_CHARACTERS + SUB_DELIMITER_CHARACTERS),
                re.escape(RESERVED_CHARACTERS + SUB_DELIMITER_CHARACTERS),
            ),
        ),
    ]
